Unit SPUnit;
{ DEFINE LOG}
Interface

Type
  PASeg = ^TASeg;
  TASeg = Array[0..8191] Of Byte;

Var
  SPBank   : Byte;
  SPSearch : Word;
  SPBoot   : Boolean;

  WaveLen : Array[1..4] Of Byte; { The fourth byte is unused in these arrays }
  Pulse   : Array[1..4] Of Byte;
  Phase   : Array[1..4] Of Byte;
  Signal  : Array[1..4] Of Byte;

Function HandleStarpath(AtariSeg: PASeg; ROMBackup: Pointer; PC,LastAddr,SPBankNum: Word;
                        TIA_Count,TIA_Scan: Integer; ROMSize: Word): Boolean;
Procedure HandlePitfall2(AtariSeg: PASeg; BSValue,Addr,XY,Acc,PC: Word);
Procedure FillDump(Const St: String);
Procedure LoadStarpath(Var F: File; P: Pointer; Size: Word);

Implementation

Uses HexWrite,OPCrt,Keyboard,AtariDbg,AtariDis;

Const
  _32BIT_OPERAND = $66;
(*
  Mnem : Array[0..255] Of String[3] =
   ('BRK','ORA','   ','slo','skb','ORA','ASL','slo',   { 00-07 }
    'PHP','ORA','ASL','anc','skw','ORA','ASL','slo',   { 08-0F }
    'BPL','ORA','   ','slo','skb','ORA','ASL','slo',   { 10-17 }
    'CLC','ORA','NOP','slo','skw','ORA','ASL','slo',   { 18-1F }
    'JSR','AND','   ','rla','BIT','AND','ROL','rla',   { 20-27 }
    'PLP','AND','ROL','anc','BIT','AND','ROL','rla',   { 28-2F }
    'BMI','AND','   ','rla','skb','AND','ROL','rla',   { 30-37 }
    'SEC','AND','NOP','rla','skw','AND','ROL','rla',   { 38-3F }
    'RTI','EOR','   ','sre','skb','EOR','LSR','sre',   { 40-47 }
    'PHA','EOR','LSR','asr','JMP','EOR','LSR','sre',   { 48-4F }
    'BVC','EOR','   ','sre','skb','EOR','LSR','sre',   { 50-57 }
    'CLI','EOR','NOP','sre','skw','EOR','LSR','sre',   { 58-5F }
    'RTS','ADC','   ','   ','skb','ADC','ROR','   ',   { 60-67 }
    'PLA','ADC','ROR','   ','JMP','ADC','ROR','   ',   { 68-6F }
    'BVS','ADC','   ','   ','skb','ADC','ROR','   ',   { 70-77 }
    'SEI','ADC','NOP','   ','skw','ADC','ROR','   ',   { 78-7F }
    'skb','STA','skb','sax','STY','STA','STX','sax',   { 80-87 }
    'DEY','NOP','TXA','ane','STY','STA','STX','sax',   { 88-8F }
    'BCC','STA','   ','sha','STY','STA','STX','sax',   { 90-97 }
    'TYA','STA','TXS','shs','shy','STA','shx','sha',   { 98-9F }
    'LDY','LDA','LDX','lax','LDY','LDA','LDX','lax',   { A0-A7 }
    'TAY','LDA','TAX','lxa','LDY','LDA','LDX','lax',   { A8-AF }
    'BCS','LDA','   ','lax','LDY','LDA','LDX','lax',   { B0-B7 }
    'CLV','LDA','TSX','las','LDY','LDA','LDX','lax',   { B8-BF }
    'CPY','CMP','skb','dcp','CPY','CMP','DEC','dcp',   { C0-C7 }
    'INY','CMP','DEX','sbx','CPY','CMP','DEC','dcp',   { C8-CF }
    'BNE','CMP','   ','dcp','skb','CMP','DEC','dcp',   { D0-D7 }
    'CLD','CMP','NOP','dcp','skw','CMP','DEC','dcp',   { D8-DF }
    'CPX','SBC','skb','   ','CPX','SBC','INC','   ',   { E0-E7 }
    'INX','SBC','NOP','   ','CPX','SBC','INC','   ',   { E8-EF }
    'BEQ','SBC','   ','   ','skb','SBC','INC','   ',   { F0-F7 }
    'SED','SBC','NOP','   ','skw','SBC','INC','   ');  { F8-FF }
*)
(*
  RW : Array[0..255] Of Char =
   (' ','R',' ','M',' ','R','M','M',' ','R','M',' ',' ','R','M','M', { 00 }
    ' ','R',' ','M',' ','R','M','M',' ','R',' ',' ',' ','R','M','M', { 10 }
    ' ','R',' ','M','R','R','M','M',' ','R','M',' ','R','R','M','M', { 20 }
    ' ','R',' ','M',' ','R','M','M',' ','R',' ',' ',' ','R','M','M', { 30 }
    ' ','R',' ','M',' ','R','M',' ',' ','R','M',' ',' ','R','M','M', { 40 }
    ' ','R',' ','M',' ','R','M',' ',' ','R',' ',' ',' ','R','M','M', { 50 }
    ' ','R',' ','M',' ','R','M',' ',' ','R','M',' ',' ','R','M','M', { 60 }
    ' ','R',' ','M',' ','R','M',' ',' ','R',' ',' ',' ','R','M','M', { 70 }
    ' ','W',' ',' ','W','W','W','W',' ',' ',' ',' ','W','W','W','W', { 80 }
    ' ','W',' ',' ','W','W','W','W',' ','W',' ',' ','W','W','W','W', { 90 }
    'R','R','R','R','R','R','R','R',' ','R',' ','R','R','R','R','R', { A0 }
    ' ','R',' ','R','R','R','R','R',' ','R',' ','R','R','R','R','R', { B0 }
    'R','R',' ','M',' ',' ','M','M',' ','R',' ','M','R','R','M','M', { C0 }
    ' ','R',' ','M',' ',' ','M','M',' ','R',' ','M',' ','R','M','M', { D0 }
    'R','R',' ',' ','R',' ','M','M',' ','R',' ',' ','R','R','M','M', { E0 }
    ' ','R',' ',' ',' ',' ','M','M',' ','R',' ',' ',' ','R','M','M');{ F0 }
*)
(*
  ATyp : Array[0..255] Of String[3] =
   ('IMP','I.X','   ','I.X','IMP','Z.P','Z.P','Z.P',   { 00-07 }
    'IMP','IMM','ACC','IMM','IMP','ABS','ABS','ABS',   { 08-0F }
    'REL','I.Y','   ','I.Y','IMP','Z.X','Z.X','Z.X',   { 10-17 }
    'IMP','A.Y','IMP','A.Y','IMP','A.X','A.X','A.X',   { 18-1F }
    'ABS','I.X','   ','I.X','Z.P','Z.P','Z.P','Z.P',   { 20-27 }
    'IMP','IMM','ACC','IMM','ABS','ABS','ABS','ABS',   { 28-2F }
    'REL','I.Y','   ','I.Y','IMP','Z.X','Z.X','Z.X',   { 30-37 }
    'IMP','A.Y','IMP','A.Y','IMP','A.X','A.X','A.X',   { 38-3F }
    'IMP','I.X','   ','I.X','IMP','Z.P','Z.P','Z.P',   { 40-47 }
    'IMP','IMM','ACC','IMM','ABS','ABS','ABS','ABS',   { 48-4F }
    'REL','I.Y','   ','I.Y','IMP','Z.X','Z.X','Z.X',   { 50-57 }
    'IMP','A.Y','IMP','A.Y','IMP','A.X','A.X','A.X',   { 58-5F }
    'IMP','I.X','   ','   ','IMP','Z.P','Z.P','   ',   { 60-67 }
    'IMP','IMM','ACC','   ','IND','ABS','ABS','   ',   { 68-6F }
    'REL','I.Y','   ','   ','IMP','Z.X','Z.X','   ',   { 70-77 }
    'IMP','A.Y','IMP','   ','IMP','A.X','A.X','   ',   { 78-7F }
    'IMP','I.X','IMP','I.X','Z.P','Z.P','Z.P','Z.P',   { 80-87 }
    'IMP','IMP','IMP','IMM','ABS','ABS','ABS','ABS',   { 88-8F }
    'REL','I.Y','   ','I.Y','Z.X','Z.X','Z.Y','Z.Y',   { 90-97 }
    'IMP','A.Y','IMP','IMP','A.X','A.X','A.Y','A.Y',   { 98-9F }
    'IMM','I.X','IMM','I.X','Z.P','Z.P','Z.P','Z.P',   { A0-A7 }
    'IMP','IMM','IMP','I+A','ABS','ABS','ABS','ABS',   { A8-AF }
    'REL','I.Y','   ','I.Y','Z.X','Z.X','Z.Y','Z.Y',   { B0-B7 }
    'IMP','A.Y','IMP','ASY','A.X','A.X','A.Y','A.Y',   { B8-BF }
    'IMM','I.X','IMP','I.X','Z.P','Z.P','Z.P','Z.P',   { C0-C7 }
    'IMP','IMM','IMP','IMM','ABS','ABS','ABS','ABS',   { C8-CF }
    'REL','I.Y','   ','I.Y','IMP','Z.X','Z.X','Z.X',   { D0-D7 }
    'IMP','A.Y','IMP','A.Y','IMP','A.X','A.X','A.X',   { D8-DF }
    'IMM','I.X','IMP','   ','Z.P','Z.P','Z.P','   ',   { E0-E7 }
    'IMP','IMM','IMP','   ','ABS','ABS','ABS','   ',   { E8-EF }
    'REL','I.Y','   ','   ','IMP','Z.X','Z.X','   ',   { F0-F7 }
    'IMP','A.Y','IMP','   ','IMP','A.X','A.Y','   ');  { F8-FF }

  Cycles : Array[0..255] Of Byte =
   (7,6,0,8,3,3,5,5, 3,2,2,2,4,4,6,6,  { 00-0F }
    2,5,0,8,4,4,6,6, 2,4,2,7,4,4,7,7,  { 10-1F }
    6,6,0,8,3,3,5,5, 4,2,2,2,4,4,6,6,  { 20-2F }
    2,5,0,8,4,4,6,6, 2,4,2,7,4,4,7,7,  { 30-3F }
    6,6,0,8,3,3,5,5, 3,2,2,2,3,4,6,6,  { 40-4F }
    2,5,0,8,4,4,6,6, 2,4,2,7,4,4,7,7,  { 50-5F }
    6,6,0,0,3,3,5,0, 4,2,2,0,5,4,6,0,  { 60-6F }
    2,5,0,0,4,4,6,0, 2,4,2,0,4,4,7,0,  { 70-7F }
    2,6,2,6,3,3,3,3, 2,2,2,2,4,4,4,4,  { 80-8F }
    2,6,0,6,4,4,4,4, 2,5,2,5,5,5,5,5,  { 90-9F }
    2,6,2,6,3,3,3,3, 2,2,2,2,4,4,4,4,  { A0-AF }
    2,5,0,5,4,4,4,4, 2,4,2,4,4,4,4,4,  { B0-BF }
    2,6,2,8,3,3,5,5, 2,2,2,2,4,4,6,6,  { C0-CF }
    2,5,0,8,4,4,6,6, 2,4,2,7,4,4,7,7,  { D0-DF }
    2,6,2,0,3,3,5,0, 2,2,2,0,4,4,6,0,  { E0-EF }
    2,5,0,0,4,4,6,0, 2,4,2,0,0,4,7,0); { F0-FF }
*)
  Cycles : Array[0..255] Of Byte =
   (7,6,0,8,3,3,5,5, 3,2,2,2,4,4,6,6,  { 00-0F }
    2,5,0,8,4,4,6,6, 2,4,2,7,4,4,7,7,  { 10-1F }
    6,6,0,8,3,3,5,5, 4,2,2,2,4,4,6,6,  { 20-2F }
    2,5,0,8,4,4,6,6, 2,4,2,7,4,4,7,7,  { 30-3F }
    6,6,0,8,3,3,5,5, 3,2,2,2,3,4,6,6,  { 40-4F }
    2,5,0,8,4,4,6,6, 2,4,2,7,4,4,7,7,  { 50-5F }
    6,6,0,8,3,3,5,5, 4,2,2,2,5,4,6,6,  { 60-6F }
    2,5,0,8,4,4,6,6, 2,4,2,7,4,4,7,7,  { 70-7F }
    2,6,2,6,3,3,3,3, 2,2,2,2,4,4,4,4,  { 80-8F }
    2,6,0,6,4,4,4,4, 2,5,2,5,5,5,5,5,  { 90-9F }
    2,6,2,6,3,3,3,3, 2,2,2,2,4,4,4,4,  { A0-AF }
    2,5,0,5,4,4,4,4, 2,4,2,4,4,4,4,4,  { B0-BF }
    2,6,2,8,3,3,5,5, 2,2,2,2,4,4,6,6,  { C0-CF }
    2,5,0,8,4,4,6,6, 2,4,2,7,4,4,7,7,  { D0-DF }
    2,6,2,8,3,3,5,5, 2,2,2,2,4,4,6,6,  { E0-EF }
    2,5,0,8,4,4,6,6, 2,4,2,7,0,4,7,7); { F0-FF }
(*
  Bytes : Array[0..255] Of Word =
   (1,2,1,2,2,2,2,2, 1,2,1,2,3,3,3,3,  { 00-0F }
    2,2,1,2,2,2,2,2, 1,3,1,3,3,3,3,3,  { 10-1F }
    3,2,1,2,2,2,2,2, 1,2,1,2,3,3,3,3,  { 20-2F }
    2,2,1,2,2,2,2,2, 1,3,1,3,3,3,3,3,  { 30-3F }
    1,2,1,2,2,2,2,2, 1,2,1,2,3,3,3,3,  { 40-4F }
    2,2,1,2,2,2,2,2, 1,3,1,3,3,3,3,3,  { 50-5F }
    1,2,1,1,2,2,2,1, 1,2,1,1,3,3,3,1,  { 60-6F }
    2,2,1,1,2,2,2,1, 1,3,1,1,3,3,3,1,  { 70-7F }
    2,2,2,2,2,2,2,2, 1,1,1,2,3,3,3,3,  { 80-8F }
    2,2,1,2,2,2,2,2, 1,3,1,3,3,3,3,3,  { 90-9F }
    2,2,2,2,2,2,2,1, 1,2,1,1,3,3,3,3,  { A0-AF }
    2,2,1,2,2,2,2,2, 1,3,1,3,3,3,3,3,  { B0-BF }
    2,2,2,1,2,2,2,2, 1,2,1,2,3,3,3,3,  { C0-CF }
    2,2,1,2,2,2,2,2, 1,3,1,3,3,3,3,3,  { D0-DF }
    2,2,2,1,2,2,2,1, 1,2,1,1,3,3,3,1,  { E0-EF }
    2,2,1,1,2,2,2,1, 1,3,1,1,3,3,3,1); { F0-FF }
*)
  SP1 : Array[0..7] Of Word = ($1000,$0000,$1000,$0000,$1000,$0800,$1000,$0800);
  SP2 : Array[0..7] Of Word = ($1800,$1800,$0000,$1000,$1800,$1800,$0800,$1000);

Type
  PPF2Dump = ^TPF2Dump;
  TPF2Dump = Array[0..2048 + 255 - 1] Of Byte;

Var
  SPWrite     : Boolean;
  SPWriteVal  : Byte;
  Listing     : Text;
  OldExitProc : Pointer;
  C0,S0       : Integer;
  C1,S1       : Integer;
  PF2Dump     : PPF2Dump;
  PF2Read     : Array[0..23] Of Integer;
  PF2ReadB    : Array[0..7] Of Boolean;
  Rand        : Array[0..3] Of Byte;
  Wave        : Array[1..3]  Of Byte;

  SPOfs       : Word;

Procedure LoadStarpath(Var F: File; P: Pointer; Size: Word);
Var
  I,J    : Word;
  Loads  : Word;
  Blocks : Word;
  Dest   : Word;
  Header : Array[0..$57] Of Byte;
  P1     : Pointer;

Begin
  Loads := Size Div 8448;
  Seek(F,0);
  BlockRead(F,P^,Size);
  For I := 0 To Loads - 1 Do
  Begin
    If I > 0 Then
     Move(Mem[Seg(P^):Ofs(P^) + (I - 1) * 8448],
          Mem[Seg(P^):Ofs(P^) + I       * 8448],8192);
    Seek(F,I * 8448 + $2000);
    BlockRead(F,Header,SizeOf(Header));
    Blocks := Header[3];
    For J := 0 To Blocks - 1 Do
    Begin
      Seek(F,I * 8448 + J * 256);
      Dest := Header[J + $10];
      Dest := ((Dest And $1C) Shl 6) + ((Dest And 3) Shl 11) + I * 8448;
      P1   := Ptr(Seg(P^),Ofs(P^) + Dest);
      BlockRead(F,P1^,256);
    End; { For J }
  End; { For I }
End; { LoadStarpath }


Function HandleStarpath(AtariSeg: PASeg; ROMBackup: Pointer; PC,LastAddr,SPBankNum: Word;
                        TIA_Count,TIA_Scan: Integer; ROMSize: Word): Boolean;
Const
{
  TIA_Name : Array[0..$2C] Of String[6] =
   ('VSYNC' ,'VBLANK','WSYNC' ,'RSYNC' ,'NUSIZ0','NUSIZ1','COLUP0','COLUP1',
    'COLUPF','COLUBK','CTRLPF','REFP0' ,'REFP1' ,'PF0'   ,'PF1'   ,'PF2',
    'RESP0' ,'RESP1' ,'RESM0' ,'RESM1' ,'RESBL' ,'AUDC0' ,'AUDC1' ,'AUDF0',
    'AUDF1' ,'AUDV0' ,'AUDV1' ,'GRP0'  ,'GRP1'  ,'ENAM0' ,'ENAM1' ,'ENABL' ,
    'HMP0'  ,'HMP1'  ,'HMM0'  ,'HMM1'  ,'HMBL'  ,'VDELP0','VDELP1','VDELBL',
    'RESMP0','RESMP1','HMOVE' ,'HMCLR' ,'CXCLR');

  PIA_Name : Array[$280..$297] Of String[8] =
   ('SWCHA','SWACNT','SWCHB','SWBCNT','INTTIM','INTTIM','','','','','','','','',
    '','','','','','','TIM1T','TIM8T','TIM64T','TIM1024T');

  TIA_Read_Name : Array[0..$D] Of String[6] =
   ('CXM0P' ,'CXM1P', 'CXP0FB','CXP1FB','CXM0FB','CXM1FB','CXBLPF','CXPPMM',
    'INPUT0','INPUT1','INPUT2','INPUT3','INPUT4','INPUT5');
}
  Boot: Array[0..$B] Of Byte =
   ($AD,$00,$0E,$A2,$FF,$A0,$00,$A9,$00,$4C,$F7,$00);

Var
  Opcode : Byte;
  Addr1  : Word;
  Addr2  : Word;
  Addr3  : Word;

  CInc1  : Integer;
  CInc2  : Integer;
  CInc3  : Integer;
  CInc4  : Integer;

  OTyp   : String[3];
  J      : Word;
  B      : Byte;

  Buf    : Pointer;
  Found  : Boolean;
  Index  : Byte;
  Start  : Word;

  Function TransWordWrite(Addr: Word): String;
  Begin
    If Addr < $80 Then Addr := Addr And $3F;
    Case Addr Of
      0..$2C: TransWordWrite := TIA_Name[Addr];
  $280..$297: TransWordWrite := PIA_Name[Addr];
  $380..$397: TransWordWrite := PIA_Name[Addr - $100];
    Else TransWordWrite := HexWord(Addr);
    End; { Case }
  End; { TransWordWrite }


  Function TransWordRead(Addr: Word): String;
  Begin
    Case Addr Of
      0..$7F: TransWordRead := TIA_Read_Name[Addr And $F];
  $280..$297: TransWordRead := PIA_Name[Addr];
  $380..$397: TransWordRead := PIA_Name[Addr - $100];
    Else TransWordRead := HexWord(Addr);
    End; { Case }
  End; { TransWordRead }


  Function TransByteRead(Addr: Word): String;
  Begin
    Case Addr Of
      0..$7F: TransByteRead := TIA_Read_Name[Addr And $F];
  $280..$297: TransByteRead := PIA_Name[Addr];
  $380..$397: TransByteRead := PIA_Name[Addr - $100];
     Else TransByteRead := HexByte(Addr);
     End; { Case }
  End; { TransByteRead }


  Function TransByteWrite(Addr: Word): String;
  Begin
    If Addr < $80 Then Addr := Addr And $3F;
    Case Addr Of
      0..$2C: TransByteWrite := TIA_Name[Addr];
    Else TransByteWrite := HexByte(Addr);
    End; { Case }
  End; { TransByteWrite }


  Function TransWord(Addr: Word): String;
  Var Instr: String[3];
  Begin
    Instr := Mnem[Opcode];
    If (Addr < $80) And (Copy(Instr,1,2) = 'RO')
     Then TransWord := TransWordRead(Addr)
     Else
      If (Instr           = 'ADC') Or
         (Instr           = 'DEC') Or
         (Instr           = 'INC') Or
         (Instr           = 'SBC') Or
         (Copy(Instr,1,2) = 'PH')  Or
         (Copy(Instr,1,2) = 'RO')  Or
         (Copy(Instr,1,2) = 'ST')
       Then TransWord := TransWordWrite(Addr)
       Else TransWord := TransWordRead(Addr);
  End; { TransWord }


  Function TransByte(Addr: Word): String;
  Var Instr: String[3];
  Begin
    Instr := Mnem[Opcode];
    If (Addr < $80) And (Copy(Instr,1,2) = 'RO')
     Then TransByte := TransByteRead(Addr)
     Else
      If (Instr           = 'ADC') Or
         (Instr           = 'DEC') Or
         (Instr           = 'INC') Or
         (Instr           = 'SBC') Or
         (Copy(Instr,1,2) = 'PH')  Or
         (Copy(Instr,1,2) = 'RO')  Or
         (Copy(Instr,1,2) = 'ST')
       Then TransByte := TransByteWrite(Addr)
       Else TransByte := TransByteRead(Addr);
  End; { TransByte }


  Procedure GetWrite;
  Begin
    If Addr3 > $10FF Then
    Begin
      Addr3 := LastAddr;
      CInc3 := CInc4;
    End;
    If Addr2 > $10FF Then
    Begin
      Addr2 := Addr3;
      CInc2 := CInc3;
    End;
    If Addr1 > $10FF Then
    Begin
      Addr1 := Addr2;
      CInc1 := CInc2;
    End;
    If (Addr1 >= $1000) And (Addr1 <= $10FF) Then
    Begin
      SPWrite    := True;
      SPWriteVal := Addr1{LastAddr} And $FF;
      C0         := TIA_Count;
      S0         := TIA_Scan;
      If S0 > 250 Then S0 := 250;
    End;
  End; { GetWrite }

Begin
  If Not SPBoot Then Move(Boot,AtariSeg^[$0E00],SizeOf(Boot));
(*
  Begin
    AtariSeg^[$0E00] := $AD;   { LDA 0E00 }
    MemW[Seg(AtariSeg^):Ofs(AtariSeg^) + $0E01] := $0E00;
    MemW[Seg(AtariSeg^):Ofs(AtariSeg^) + $0E03] := $FFA2; { LDX #FF  }
    MemW[Seg(AtariSeg^):Ofs(AtariSeg^) + $0E05] := $00A0; { LDY #00  }
    MemW[Seg(AtariSeg^):Ofs(AtariSeg^) + $0E07] := $00A9; { LDA #00  }
    Mem[Seg(AtariSeg^):Ofs(AtariSeg^)  + $0E09] := $4C;   { JMP 00F7 }
    MemW[Seg(AtariSeg^):Ofs(AtariSeg^) + $0E0A] := $00F7;
    SPBoot := True;
  End;
*)
  SPBank         := SPBankNum;
  HandleStarpath := False;
  Opcode         := AtariSeg^[PC];

  If (LastAddr = $0E00) And (PC = $0E00) Then
  Begin
    B        := AtariSeg^[$FA];
    Found    := False;
    SPSearch := 0;
    While (SPSearch < ROMSize) And (SPSearch < 4 * 8448) And Not Found Do
    Begin
      Index := Mem[Seg(ROMBackup^):Ofs(ROMBackup^) + SPSearch + $2005];
      If Index = B Then Found := True Else Inc(SPSearch,8448);
    End; { While }
    If Found Then
    Begin

      FillChar(AtariSeg^[$81],$1D,#0);
      B := Mem[Seg(ROMBackup^):Ofs(ROMBackup^) + SPSearch + $2002];
      AtariSeg^[$80] := B;
      AtariSeg^[$F7] := $CD;   { CMP 1000 }
      AtariSeg^[$F8] := $00;
      AtariSeg^[$F9] := $10;
      AtariSeg^[$FA] := $AD;   { LDA 1FF8 }
      AtariSeg^[$FB] := $F8;
      AtariSeg^[$FC] := $1F;
      Start := MemW[Seg(ROMBackup^):Ofs(ROMBackup^) + SPSearch + $2000];
      AtariSeg^[$FD] := $4C;   { JMP Start }
      MemW[Seg(AtariSeg^):Ofs(AtariSeg^) + $FE] := Start;
      AtariSeg^[$F8]  := B;

      SPBankNum := B;
      Move(Mem[Seg(ROMBackup^):Ofs(ROMBackup^) + SP1[(SPBankNum Shr 2) And 7] + SPSearch],
           AtariSeg^[$1000],$800);
      Move(Mem[Seg(ROMBackup^):Ofs(ROMBackup^) + SP2[(SPBankNum Shr 2) And 7] + SPSearch],
           AtariSeg^[$1800],$800);

{      AtariSeg^[$1FF8] := B;}

      SPWrite        := False;
      SPBank         := SPBankNum;
      HandleStarpath := True;
    End;
    Exit;
  End;


{
  Inc(TIA_Count,3 * Cycles[Opcode]);
  If TIA_Count >= 160 Then
  Begin
    Dec(TIA_Count,228);
    If TIA_Scan < 250 Then Inc(TIA_Scan);
  End;
}

  { Log the instruction }
{$IFDEF LOG}
  Write(Listing,HexWord(PC),'    ',Mnem[Opcode],'   ');
  OTyp := ATyp[Opcode];
  If OTyp = 'ABS' Then Write(Listing,'   ',TransWord(MemW[Seg(AtariSeg^):Ofs(AtariSeg^) + PC + 1] And $1FFF));
  If OTyp = 'A.X' Then Write(Listing,'   ',TransWord(MemW[Seg(AtariSeg^):Ofs(AtariSeg^) + PC + 1] And $1FFF),',X');
  If OTyp = 'A.Y' Then Write(Listing,'   ',TransWord(MemW[Seg(AtariSeg^):Ofs(AtariSeg^) + PC + 1] And $1FFF),',Y');
  If OTyp = 'ASY' Then Write(Listing,'   (',TransWord(MemW[Seg(AtariSeg^):Ofs(AtariSeg^) + PC + 1] And $1FFF),',Y) & S');
  If OTyp = 'IND' Then Write(Listing,'   (',HexWord(MemW[Seg(AtariSeg^):Ofs(AtariSeg^) + PC + 1] And $1FFF),')');
  If OTyp = 'I.X' Then Write(Listing,'   (',HexByte(AtariSeg^[PC + 1]),',X)');
  If OTyp = 'I.Y' Then Write(Listing,'   (',HexByte(AtariSeg^[PC + 1]),'),Y');
  If OTyp = 'ACC' Then Write(Listing,'   A');
  If OTyp = 'IMM' Then Write(Listing,'   #',HexByte(AtariSeg^[PC + 1]));
  If OTyp = 'Z.P' Then Write(Listing,'   ',TransByte(AtariSeg^[PC + 1]));
  If OTyp = 'Z.X' Then Write(Listing,'   ',TransByte(AtariSeg^[PC + 1]),',X');
  If OTyp = 'Z.Y' Then Write(Listing,'   ',TransByte(AtariSeg^[PC + 1]),',Y');
  If OTyp = 'I+A' Then Write(Listing,'   #',HexByte(AtariSeg^[PC + 1]),' + A');
  If OTyp = 'REL' Then
  Begin
    B := AtariSeg^[PC + 1];
    Asm
      MOV   BX,WORD PTR PC
      ADD   BX,2
      MOV   AL,BYTE PTR B
      CBW
      ADD   BX,AX
      MOV   WORD PTR J,BX
    End; { Asm }
    Write(Listing,'   ',HexWord(J));
  End;
  If SPWrite Then Write(Listing,'    ',SPWrite,'  (',HexByte(SPWriteVal),')');
{  WriteLn(Listing);}
{$ENDIF}

  { Get all the addresses that can be accessed }

  If (ATyp[Opcode] = 'A.X') Or           { Absolute X }
     (ATyp[Opcode] = 'A.Y') Or           { Absolute Y }
     (ATyp[Opcode] = 'I.Y') Then         { Indirect Y }
  Begin
    SPOfs := (LastAddr - (MemW[Seg(AtariSeg^):Ofs(AtariSeg^) + PC + 1] And $1FFF)) And $FF;

    If Copy(ATyp[Opcode],1,1) = 'A' Then
    Begin
      Addr1 := MemW[Seg(AtariSeg^):Ofs(AtariSeg^) + PC + 1] And $1FFF;
      Addr3 := (Addr1 And $FF00) + (LastAddr And $FF);
      Addr2 := Addr3;
      Addr1 := Addr3;

      If Opcode In [$19,$1D,$39,$3D,$59,$5D,$79,$7D,$B9,$BB,$BC,$BD,$BE,$BF,$D9,$DD,$F9,$FD] Then
      Begin
        CInc1 := 4;                      { Read }
        CInc2 := 4;
        CInc3 := 4;
        CInc4 := 4;
        If (Addr3 And $FF00) <> (LastAddr And $FF00) Then
        Begin
          Inc(CInc1);
          Inc(CInc2);
          Inc(CInc3);
          Inc(CInc4);
        End;
      End
      Else
       If Opcode In [$99,$9C,$9D,$9E,$9F] Then
       Begin
         CInc1 := 4;                     { Write }
         CInc2 := 4;
         CInc3 := 4;
         CInc4 := 5;
       End
       Else
       Begin
         CInc1 := 4;                     { Read/Modify/Write }
         CInc2 := 5;
         CInc3 := 6;
         CInc4 := 7;
         Addr2 := LastAddr;
         Addr3 := LastAddr;
       End;
    End
    Else
    Begin
{
      Addr1 := MemW[Seg(AtariSeg^):Ofs(AtariSeg^) + PC + 1] And $1FFF;
      Addr2 := MemW[Seg(AtariSeg^):Ofs(AtariSeg^) + Addr1] And $1FFF;
      Addr3 := (Addr2 And $FF00) + (LastAddr And $FF);
      Addr2 := Addr3;
}
      Addr1 := MemW[Seg(AtariSeg^):Ofs(AtariSeg^) + PC + 1] And $FF;
      Addr2 := MemW[Seg(AtariSeg^):Ofs(AtariSeg^) + Addr1] And $1FFF;
      Addr3 := (Addr2 And $FF00) + (LastAddr And $FF);
      Addr1 := Addr3;
      Addr2 := Addr3;

      If Opcode In [$11,$31,$51,$71,$B1,$D1,$F1] Then  
      Begin
        CInc1 := 5;                      { Read }
        CInc2 := 5;
        CInc3 := 5;
        CInc4 := 5;
        If (Addr3 And $FF00) <> (LastAddr And $FF00) Then
        Begin
          Inc(CInc1);
          Inc(CInc2);
          Inc(CInc3);
          Inc(CInc4);
        End;
      End
      Else
       If Opcode In [$91,$93] Then                     
       Begin
         CInc1 := 5;                     { Write }
         CInc2 := 5;
         CInc3 := 5;
         CInc4 := 6;
       End
       Else
       Begin                                           
         CInc1 := 5;                     { Read/Modify/Write }
         CInc2 := 6;
         CInc3 := 7;
         CInc4 := 8;
         Addr2 := LastAddr;
         Addr3 := LastAddr;
       End;
    End;
  End
  Else
  Begin
    CInc4 := Cycles[Opcode];
    CInc1 := CInc4;
    CInc2 := CInc4;
    CInc3 := CInc4;
    Addr1 := LastAddr;                   { Absolute/Indirect X }
    Addr2 := LastAddr;
    Addr3 := LastAddr;
  End;

  CInc1 := CInc1 * 3;
  CInc2 := CInc2 * 3;
  CInc3 := CInc3 * 3;
  CInc4 := CInc4 * 3;

  { Get rid of any addresses that are below the RAM range }

  If Addr3 < $1000 Then
  Begin
    Addr3 := LastAddr;
    CInc3 := CInc4;
  End;
  If Addr2 < $1000 Then
  Begin
    Addr2 := Addr3;
    CInc2 := CInc3;
  End;
  If Addr1 < $1000 Then
  Begin
    Addr1 := Addr2;
    CInc1 := CInc2;
  End;

  Inc(TIA_Count,CInc1);
  If TIA_Count >= 160 Then
  Begin
    Dec(TIA_Count,228);
{    If TIA_Scan < 250 Then} Inc(TIA_Scan);
  End;

  { Check to see if we already have the value to write }
(*
  TextMode(C80);
  WriteLn('SPWrite=',SPWrite);
  WriteLn('Addr1=',HexWord(Addr1));
  WriteLn;
  WriteLn('DSeg=',HexWord(DSeg));
  WriteLn('@SPWrite=',HexWord(Seg(SPWrite)),':',HexWord(Ofs(SPWrite)));
  Key[kSPACE] := 0;
  Repeat Until Key[kSPACE] <> 0;
  Asm
    MOV   AX,13h
    INT   10h
  End; { Asm }
*)



  If SPWrite Then
  Begin
    SPWrite := False;

    { If writing isn't enabled and it's not a bank switch, save this as a }
    { new write value intead if possible                                  }


    C1 := TIA_Count;
    S1 := TIA_Scan;
    If TIA_Scan > S0 Then Inc(C1,228{160} * (TIA_Scan - S0));


    If  (TIA_Scan < S0) Or
        ((TIA_Scan = S0) And (C1 < C0 + 12)) Or
        (C1 > C0 + 18) Or
{       (Addr1 < $1100) Or}

    (   (Addr1 <> $1FF8) And

       (((SPBankNum And 2) = 0) Or
        ((Addr1 >= $1800) And ((SPBankNum And 8) = 0))) )  Then
    Begin

      { Find the first address that accesses 1000-10FF }

      GetWrite;
{$IFDEF LOG}
      WriteLn(Listing);
{$ENDIF}
    End
    Else
    Begin
{
      If TIA_Scan = S0 + 1 Then Inc(TIA_Count,160);
      If TIA_Count <= C0 + 3 * 6 Then Addr1 := LastAddr;
}
{
      If ((Addr1 And $FF) = (LastAddr And $FF)) And
         ((Addr1 And $FF00) = (LastAddr And $FF00) - $100) Then Inc(Addr1,$100);
}


      Addr1 := LastAddr;



      { Write to the location }
{$IFDEF LOG}
      WriteLn(Listing,'   * --> ',HexWord(Addr1) + ' Last= ' + HexWord(LastAddr));
{$ENDIF}
      If Addr1 <> $1FF8 Then AtariSeg^[Addr1] := SPWriteVal;
      If Addr1 = $1FF8 Then
      Begin
        Move(AtariSeg^[$1000],
             Mem[Seg(ROMBackup^):Ofs(ROMBackup^) + SP1[(SPBankNum Shr 2) And 7] + SPSearch],$800);
        Move(AtariSeg^[$1800],
             Mem[Seg(ROMBackup^):Ofs(ROMBackup^) + SP2[(SPBankNum Shr 2) And 7] + SPSearch],$800);

        SPBankNum := SPWriteVal;
        Move(Mem[Seg(ROMBackup^):Ofs(ROMBackup^) + SP1[(SPBankNum Shr 2) And 7] + SPSearch],
             AtariSeg^[$1000],$800);
        Move(Mem[Seg(ROMBackup^):Ofs(ROMBackup^) + SP2[(SPBankNum Shr 2) And 7] + SPSearch],
             AtariSeg^[$1800],$800);

        { If the ROM bank was loaded, fill it in with my routine }

        If SP2[(SPBankNum Shr 2) And 7] = $1800 Then
        Begin
          AtariSeg^[$1800] := $4C;   { JMP 0E00 }
          AtariSeg^[$1801] := $00;
          AtariSeg^[$1802] := $0E;
        End;

        HandleStarpath := True;
        SPBank         := SPBankNum;

        If Addr1 <> $1FF8 Then AtariSeg^[Addr1] := SPWriteVal;


{$IFDEF LOG}
        WriteLn(Listing);
{$ENDIF}

      End;
    End;
  End
  Else
  Begin
{$IFDEF LOG}
    WriteLn(Listing);
{$ENDIF}
    { Find the first address that accesses 1000-10FF }

    GetWrite;
  End;

{  Move(SPWrite,Mem[Seg(AtariSeg^):16384 + 256 + Ofs(SPWrite)],Ofs(S1) - Ofs(SPWrite) + SizeOf(S1) - 1);}


End; { HandleStarpath }

{$F+}
Procedure NewExitProc;
Begin
  ExitProc := OldExitProc;
  Flush(Listing);
  Close(Listing);
  FreeMem(PF2Dump,2048 + 255);
End; { NewExitProc }
{$F-}

Procedure HandlePitfall2(AtariSeg: PASeg; BSValue,Addr,XY,Acc,PC: Word);
Var
  B       : Byte;
  I,W     : Word;
  Opcode  : Byte;
  Opcode1 : Byte;
  OTyp    : String[3];
  X       : Word;
  Y       : Word;
  Value   : Integer;
  M       : String[3];

  Function GetSound: Byte;
  Type Ar = Array[0..$7F] Of Byte;
  Var
{    _Wave   : Array[1..3] Of Word;}
    P       : ^Ar;
    Amp     : Byte;

  Begin
    P := Pointer(LongInt(AtariSeg) + $1000);

    Asm
      LES   DI,DWORD PTR P

      DB    _32BIT_OPERAND
      MOV   AX,WORD PTR [ES:DI + 45h]
      DB    _32BIT_OPERAND
      MOV   BX,WORD PTR [ES:DI + 4Dh]
      DB    _32BIT_OPERAND
      MOV   CX,WORD PTR [ES:DI + 55h]
      DB    _32BIT_OPERAND
      MOV   DX,WORD PTR [ES:DI + 5Dh]

      DB    _32BIT_OPERAND
      MOV   WORD PTR WaveLen,AX
      DB    _32BIT_OPERAND
      MOV   WORD PTR Pulse,BX
      DB    _32BIT_OPERAND
      MOV   WORD PTR Phase,CX

      DB    _32BIT_OPERAND
      SHR   DX,4
      AND   DX,0303h
      MOV   WORD PTR Signal,DX
      DB    _32BIT_OPERAND
      SHR   DX,16
      AND   DX,0303h
      MOV   WORD PTR Signal[2],DX

    End; { Asm }

{
    WaveLen[1] := P^[$45];
    WaveLen[2] := P^[$46];
    WaveLen[3] := P^[$47];

    Pulse[1]   := P^[$4D];
    Pulse[2]   := P^[$4E];
    Pulse[3]   := P^[$4F];
}
    If Pulse[1] > WaveLen[1] Then Pulse[1] := 0;
    If Pulse[2] > WaveLen[2] Then Pulse[2] := 0;
    If Pulse[3] > WaveLen[3] Then Pulse[3] := 0;
{
    Phase[1]   := P^[$55];
    Phase[2]   := P^[$56];
    Phase[3]   := P^[$57];

    Signal[1]  := (P^[$5D] Shr 4) And 3;
    Signal[2]  := (P^[$5E] Shr 4) And 3;
    Signal[3]  := (P^[$5F] Shr 4) And 3;
}
    Amp := 0;

(*
    Inc(Wave[1]);
    Inc(Wave[2]);
    Inc(Wave[3]);

    _Wave[1] := Wave[1] + Phase[1];
    _Wave[2] := Wave[2] + Phase[2];
    _Wave[3] := Wave[3] + Phase[3];

    If WaveLen[1] > 0
     Then _Wave[1] := _Wave[1] Mod WaveLen[1]
     Else _Wave[1] := 0;

    If WaveLen[2] > 0
     Then _Wave[2] := _Wave[2] Mod WaveLen[2]
     Else _Wave[2] := 0;

    If WaveLen[3] > 0
     Then _Wave[3] := _Wave[3] Mod WaveLen[3]
     Else _Wave[3] := 0;

    Case Signal[1] Of
      1: Amp := 4;
      3: If _Wave[1] <= Pulse[1] Then Amp := 4;
    End; { Case }

    Case Signal[2] Of
      1: Inc(Amp,5);
      3: If _Wave[2] <= Pulse[2] Then Inc(Amp,5);
    End; { Case }

    Case Signal[3] Of
      1: Inc(Amp,6);
      3: If _Wave[3] <= Pulse[3] Then Inc(Amp,6);
    End; { Case }
*)
    GetSound := Amp;
  End; { GetSound }

Begin
  If BSValue = $CE Then
  Begin
    If (Addr >= $1008) And (Addr <= $103F) Then
    Begin
      I := Addr And 7;
      If Addr >= $1038 Then
      Begin
        B := $FF;
        Dec(PF2Read[I]);
      End
      Else
       If Addr >= $1028 Then
       Begin
         B := PF2Dump^[(PF2Read[I] And $7FF)];
         Dec(PF2Read[I]);
       End
       Else
        If Addr >= $1020 Then
        Begin
          B := PF2Dump^[(PF2Read[I] And $7FF)];
          Asm
            MOV  AL,B
            RCR  AL,1
            RCL  AH,1
            RCR  AL,1
            RCL  AH,1
            RCR  AL,1
            RCL  AH,1
            RCR  AL,1
            RCL  AH,1
            RCR  AL,1
            RCL  AH,1
            RCR  AL,1
            RCL  AH,1
            RCR  AL,1
            RCL  AH,1
            RCR  AL,1
            RCL  AH,1
            MOV  B,AH
          End; { Asm }
          Dec(PF2Read[I]);
        End
        Else
         If Addr >= $1018 Then
         Begin
           B := PF2Dump^[(PF2Read[I] And $7FF)];
           B := (B Shr 4) + (B Shl 4);
           Dec(PF2Read[I]);
         End
         Else
          If Addr >= $1010 Then
          Begin
            B := PF2Dump^[(PF2Read[I] And $7FF)];
            If PF2Read[I] = PF2Read[I + 8]  Then PF2ReadB[I] := True;
            If PF2Read[I] = PF2Read[I + 16] Then PF2ReadB[I] := False;
            If Not PF2ReadB[I] Then B := 0;
            Dec(PF2Read[I]);
          End
          Else
           If Addr >= $1008 Then
           Begin
             B := PF2Dump^[(PF2Read[I] And $7FF)];
             Dec(PF2Read[I]);
           End;
      AtariSeg^[Addr]        := B;
      AtariSeg^[Addr + $800] := B;
      PF2Read[I + 8]  := (PF2Read[I + 8]  And $FF) + (PF2Read[I] And $700);
      PF2Read[I + 16] := (PF2Read[I + 16] And $FF) + (PF2Read[I] And $700);
    End
    Else
     Begin
       If Addr = $1006
        Then AtariSeg^[$1006] := GetSound
        Else
         If (Addr >= $1000) And (Addr <= $1003) Then
         Begin
           AtariSeg^[Addr] :=
            PF2Dump^[2048 + Rand[Addr - $1000]];
           Inc(Rand[Addr - $1000]);
           If Rand[Addr - $1000] = $255 Then Rand[Addr - $1000] := 0;
         End;
     End;
  End
  Else
   If BSValue = $CF Then
   Begin
     If (Addr >= $1040) And (Addr <= $105F){ And ((Addr And 7) < 5)} Then
     Begin
       Opcode := AtariSeg^[PC];
       X      := XY And $FF;
       Y      := XY Shr 8;
       Value  := -1;
       M      := Mnem[Opcode];
       Asm
         MOV   AX,WORD PTR M[1]
         MOV   WORD PTR W,AX
         MOV   AL,BYTE PTR M[3]
         MOV   BYTE PTR B,AL
       End; { Asm }
       If W = $5453 Then      { "ST" }
       Begin
         If B = 65 Then Value := Acc Else  { "A" }
          If B = 88 Then Value := X Else   { "X" }
           If B = 89 Then Value := Y;      { "Y" }
  {     If Mnem[Opcode] = 'STA' Then Value := Acc Else
        If Mnem[Opcode] = 'STX' Then Value := X Else
         If Mnem[Opcode] = 'STY' Then Value := Y;}
         If Value >= 0 Then
         Begin
           AtariSeg^[Addr] := Value;

           If (Addr And 7) < 5 Then
           Begin
             I := Addr And 7;
             If (Addr And $10) <> 0 Then
             Begin
               If (Addr And 8) = 0
                Then PF2Read[I] := (PF2Read[I] And $700) + Value
                Else
                Begin
                  PF2Read[I]      := (PF2Read[I]      And $FF)   + ((Value Shl 8) And $700);
                  PF2Read[I + 8]  := (PF2Read[I + 8]  And $FF)   + ((Value Shl 8) And $700);
                  PF2Read[I + 16] := (PF2Read[I + 16] And $FF)   + ((Value Shl 8) And $700);
                  PF2ReadB[I] := False;
                End;
             End
             Else
             Begin
               If (Addr And 8) = 0
                Then PF2Read[I + 8]  := (PF2Read[I + 8]  And $700) + Value
                Else PF2Read[I + 16] := (PF2Read[I + 16] And $700) + Value;
             End;
           End;
         End;
       End;
     End;
   End;
End; { HandlePitfall2 }


Procedure LoadDump;
Var F: File;
  D : Array[0..2047] Of Byte;
  I,J : Integer;

Begin
  GetMem(PF2Dump,2048 + 255);
(*
  If Exist('PIT2DATA.BIN') Then
  Begin
    Assign(F,'PIT2DATA.BIN');
    Reset(F,1);
    BlockRead(F,D,2048);
    BlockRead(F,PF2Dump^[2048],255);
    Close(F);
    J := 2047;
    For I := 0 To 2047 Do
    Begin
      PF2Dump^[I] := D[J];
      Dec(J);
      If J < 0 Then J := 2047;
    End; { For I }
  End;
*)
End; { LoadDump }


Procedure FillDump(Const St: String);
Var
  F   : File;
  D   : Array[0..2047] Of Byte;
  I,J : Integer;

Begin
  If Exist(St) Then
  Begin
    Assign(F,St);
    Reset(F,1);
    Seek(F,8192);
    BlockRead(F,D,2048);
    BlockRead(F,PF2Dump^[2048],255);
    Close(F);
    J := 2047;
    For I := 0 To 2047 Do
    Begin
      PF2Dump^[I] := D[J];
      Dec(J);
      If J < 0 Then J := 2047;
    End; { For I }
  End;
End; { FillDump }


Begin
  SPSearch   := 0;
  SPWrite    := False;
  SPWriteVal := 0;
  SPBoot     := False;
{$IFDEF LOG }
  Assign(Listing,'STARPATH.LOG');
  ReWrite(Listing);
  OldExitProc := ExitProc;
  ExitProc    := @NewExitProc;
{$ENDIF}
  LoadDump;
  FillChar(Wave,SizeOf(Wave),#0);
  FillChar(Rand,SizeOf(Rand),#0);
  FillChar(PF2ReadB,SizeOf(PF2ReadB),#0);
End.